@@ -24,25 +24,36 @@
 
 /* local vars */
 static int sublevel = 0;
+vector keywords = NULL;
 
-void
-keyword_alloc(vector keywords, char *string, void (*handler) (vector))
+int
+keyword_alloc(vector keywords, char *string, int (*handler) (vector),
+		int (*print) (char *, int, void *))
 {
 	struct keyword *keyword;
 
-	vector_alloc_slot(keywords);
+	keyword = (struct keyword *) MALLOC(sizeof (struct keyword));
 
-	keyword = (struct keyword *) zalloc(sizeof (struct keyword));
+	if (!keyword)
+		return 1;
+	
+	if (!vector_alloc_slot(keywords)) {
+		FREE(keyword);
+		return 1;
+	}
 	keyword->string = string;
 	keyword->handler = handler;
+	keyword->print = print;
 
 	vector_set_slot(keywords, keyword);
+
+	return 0;
 }
 
-void
-install_keyword_root(char *string, void (*handler) (vector))
+int
+install_keyword_root(char *string, int (*handler) (vector))
 {
-	keyword_alloc(keywords, string, handler);
+	return keyword_alloc(keywords, string, handler, NULL);
 }
 
 void
@@ -57,8 +68,9 @@ install_sublevel_end(void)
 	sublevel--;
 }
 
-void
-install_keyword(char *string, void (*handler) (vector))
+int
+install_keyword(char *string, int (*handler) (vector),
+		int (*print) (char *, int, void *))
 {
 	int i = 0;
 	struct keyword *keyword;
@@ -75,8 +87,11 @@ install_keyword(char *string, void (*handler) (vector))
 	if (!keyword->sub)
 		keyword->sub = vector_alloc();
 
+	if (!keyword->sub)
+		return 1;
+
 	/* add new sub keyword */
-	keyword_alloc(keyword->sub, string, handler);
+	return keyword_alloc(keyword->sub, string, handler, print);
 }
 
 void
@@ -89,11 +104,78 @@ free_keywords(vector keywords)
 		keyword = VECTOR_SLOT(keywords, i);
 		if (keyword->sub)
 			free_keywords(keyword->sub);
-		free(keyword);
+		FREE(keyword);
 	}
 	vector_free(keywords);
 }
 
+struct keyword *
+find_keyword(vector v, char * name)
+{
+	struct keyword *keyword;
+	int i;
+	int len;
+
+	if (!name || !keywords)
+		return NULL;
+
+	if (!v)
+		v = keywords;
+
+	len = strlen(name);
+
+	for (i = 0; i < VECTOR_SIZE(v); i++) {
+		keyword = VECTOR_SLOT(v, i);
+		if ((strlen(keyword->string) == len) &&
+		    !strcmp(keyword->string, name))
+			return keyword;
+		if (keyword->sub) {
+			keyword = find_keyword(keyword->sub, name);
+			if (keyword)
+				return keyword;
+		}
+	}
+	return NULL;
+}
+
+int
+snprint_keyword(char *buff, int len, char *fmt, struct keyword *kw, void *data)
+{
+	int r;
+	int fwd = 0;
+	char *f = fmt;
+
+	if (!kw || !kw->print)
+		return 0;
+
+	do {
+		if (fwd == len || *f == '\0')
+			break;
+		if (*f != '%') {
+			*(buff + fwd) = *f;
+			fwd++;
+			continue;
+		}
+		f++;
+		switch(*f) {
+		case 'k':
+			fwd += snprintf(buff + fwd, len - fwd, kw->string);
+			break;
+		case 'v':
+			r = kw->print(buff + fwd, len - fwd, data);
+			if (!r) { /* no output if no value */
+				buff = '\0';
+				return 0;
+			}
+			fwd += r;
+			break;
+		}
+		if (fwd > len)
+			fwd = len;
+	} while (*f++);
+	return fwd;
+}
+
 vector
 alloc_strvec(char *string)
 {
@@ -122,12 +204,22 @@ alloc_strvec(char *string)
 	/* Create a vector and alloc each command piece */
 	strvec = vector_alloc();
 
+	if (!strvec)
+		return NULL;
+
 	in_string = 0;
 	while (1) {
+		if (!vector_alloc_slot(strvec))
+			goto out;
+
 		start = cp;
 		if (*cp == '"') {
 			cp++;
-			token = zalloc(2);
+			token = MALLOC(2);
+
+			if (!token)
+				goto out;
+
 			*(token) = '"';
 			*(token + 1) = '\0';
 			if (in_string)
@@ -140,13 +232,14 @@ alloc_strvec(char *string)
 				!= '\0' && *cp != '"')
 				cp++;
 			strlen = cp - start;
-			token = zalloc(strlen + 1);
+			token = MALLOC(strlen + 1);
+
+			if (!token)
+				goto out;
+
 			memcpy(token, start, strlen);
 			*(token + strlen) = '\0';
 		}
-
-		/* Alloc & set the slot */
-		vector_alloc_slot(strvec);
 		vector_set_slot(strvec, token);
 
 		while (isspace((int) *cp) && *cp != '\0')
@@ -154,6 +247,9 @@ alloc_strvec(char *string)
 		if (*cp == '\0' || *cp == '!' || *cp == '#')
 			return strvec;
 	}
+out:
+	vector_free(strvec);
+	return NULL;
 }
 
 int
@@ -183,7 +279,14 @@ read_value_block(void)
 	vector vec = NULL;
 	vector elements = vector_alloc();
 
-	buf = (char *) zalloc(MAXBUF);
+	buf = (char *) MALLOC(MAXBUF);
+
+	if (!buf)
+		return NULL;
+
+	if (!elements)
+		goto out;
+
 	while (read_line(buf, MAXBUF)) {
 		vec = alloc_strvec(buf);
 		if (vec) {
@@ -196,28 +299,39 @@ read_value_block(void)
 			if (VECTOR_SIZE(vec))
 				for (i = 0; i < VECTOR_SIZE(vec); i++) {
 					str = VECTOR_SLOT(vec, i);
-					dup = (char *) zalloc(strlen(str) + 1);
+					dup = (char *) MALLOC(strlen(str) + 1);
 					memcpy(dup, str, strlen(str));
-					vector_alloc_slot(elements);
+
+					if (!vector_alloc_slot(elements))
+						goto out1;
+
 					vector_set_slot(elements, dup);
 				}
 			free_strvec(vec);
 		}
 		memset(buf, 0, MAXBUF);
 	}
-
-	free(buf);
+	FREE(buf);
 	return elements;
+out1:
+	FREE(dup);
+out:
+	FREE(buf);
+	return NULL;
 }
 
-void
+int
 alloc_value_block(vector strvec, void (*alloc_func) (vector))
 {
 	char *buf;
 	char *str = NULL;
 	vector vec = NULL;
 
-	buf = (char *) zalloc(MAXBUF);
+	buf = (char *) MALLOC(MAXBUF);
+
+	if (!buf)
+		return 1;
+
 	while (read_line(buf, MAXBUF)) {
 		vec = alloc_strvec(buf);
 		if (vec) {
@@ -234,7 +348,8 @@ alloc_value_block(vector strvec, void (*alloc_func) (vector))
 		}
 		memset(buf, 0, MAXBUF);
 	}
-	free(buf);
+	FREE(buf);
+	return 0;
 }
 
 void *
@@ -253,11 +368,11 @@ set_value(vector strvec)
 			len += strlen(str);
 			if (!alloc)
 				alloc =
-				    (char *) zalloc(sizeof (char *) *
+				    (char *) MALLOC(sizeof (char *) *
 						    (len + 1));
 			else {
 				alloc =
-				    realloc(alloc, sizeof (char *) * (len + 1));
+				    REALLOC(alloc, sizeof (char *) * (len + 1));
 				tmp = VECTOR_SLOT(strvec, i-1);
 				if (*str != '"' && *tmp != '"')
 					strncat(alloc, " ", 1);
@@ -267,7 +382,7 @@ set_value(vector strvec)
 				strncat(alloc, str, strlen(str));
 		}
 	} else {
-		alloc = zalloc(sizeof (char *) * (size + 1));
+		alloc = MALLOC(sizeof (char *) * (size + 1));
 		memcpy(alloc, str, size);
 	}
 	return alloc;
@@ -275,23 +390,27 @@ set_value(vector strvec)
 
 /* non-recursive configuration stream handler */
 static int kw_level = 0;
-void
+int
 process_stream(vector keywords)
 {
 	int i;
+	int r = 0;
 	struct keyword *keyword;
 	char *str;
 	char *buf;
 	vector strvec;
 
-	buf = zalloc(MAXBUF);
+	buf = MALLOC(MAXBUF);
+
+	if (!buf)
+		return 1;
+
 	while (read_line(buf, MAXBUF)) {
 		strvec = alloc_strvec(buf);
 		memset(buf,0, MAXBUF);
 
-		if (!strvec) {
+		if (!strvec)
 			continue;
-		}
 
 		str = VECTOR_SLOT(strvec, 0);
 
@@ -305,11 +424,11 @@ process_stream(vector keywords)
 
 			if (!strcmp(keyword->string, str)) {
 				if (keyword->handler)
-					(*keyword->handler) (strvec);
+					r += (*keyword->handler) (strvec);
 
 				if (keyword->sub) {
 					kw_level++;
-					process_stream(keyword->sub);
+					r += process_stream(keyword->sub);
 					kw_level--;
 				}
 				break;
@@ -319,19 +438,24 @@ process_stream(vector keywords)
 		free_strvec(strvec);
 	}
 
-	free(buf);
-	return;
-
+	FREE(buf);
+	return r;
 }
 
 /* Data initialization */
-void
-init_data(char *conf_file, vector (*init_keywords) (void))
+int
+init_data(char *conf_file, void (*init_keywords) (void))
 {
+	int r;
+
+	if (!keywords)
+		keywords = vector_alloc();
+	if (!keywords)
+		return 1;
 	stream = fopen(conf_file, "r");
 	if (!stream) {
 		syslog(LOG_WARNING, "Configuration file open problem");
-		return;
+		return 1;
 	}
 
 	/* Init Keywords structure */
@@ -343,7 +467,9 @@ init_data(char *conf_file, vector (*init_keywords) (void))
 */
 
 	/* Stream handling */
-	process_stream(keywords);
+	r = process_stream(keywords);
 	fclose(stream);
-	free_keywords(keywords);
+	//free_keywords(keywords);
+
+	return r;
 }
