@@ -19,29 +19,13 @@
 #ifndef _MAIN_H
 #define _MAIN_H
 
-/*
- * configurator actions
- */
-#define ACT_NOTHING_STR		"unchanged"
-#define ACT_RELOAD_STR		"reload"
-#define ACT_SWITCHPG_STR	"switchpg"
-#define ACT_CREATE_STR		"create"
-
-enum actions {
-	ACT_RESERVED,
-	ACT_NOTHING,
-	ACT_RELOAD,
-	ACT_SWITCHPG,
-	ACT_CREATE
-};
-
 /*
  * Build version
  */
 #define PROG    "multipath"
 
-#define VERSION_CODE 0x000402
-#define DATE_CODE    0x170105
+#define VERSION_CODE 0x000407
+#define DATE_CODE    0x030c06
 
 #define MULTIPATH_VERSION(version)	\
 	(version >> 16) & 0xFF,		\
